/*
 * $Header: /home/cvs/jakarta-slide/src/share/org/apache/slide/lock/LockImpl.java,v 1.17 2001/07/24 06:42:20 remm Exp $
 * $Revision: 1.17 $
 * $Date: 2001/07/24 06:42:20 $
 *
 * ====================================================================
 *
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:  
 *       "This product includes software developed by the 
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Tomcat", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written 
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * [Additional notices, if required by prior licensing conditions]
 *
 */ 

package org.apache.slide.lock;

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;
import java.util.Date;
import java.util.Stack;
import org.apache.slide.common.*;
import org.apache.slide.structure.*;
import org.apache.slide.security.*;
import org.apache.slide.util.Configuration;
import org.apache.slide.authenticate.CredentialsToken;

/**
 * Lock helper class.
 * 
 * @author <a href="mailto:remm@apache.org">Remy Maucherat</a>
 * @version $Revision: 1.17 $
 */
public final class LockImpl implements Lock {
    
    // ----------------------------------------------------------- Constructors
    
    
    /**
     * Constructor.
     * 
     * @param namespace Associated namespace
     * @param namespaceConfig Namespace configuration
     * @param securityHelper Security helper
     */
    public LockImpl(Namespace namespace, NamespaceConfig namespaceConfig, 
                    Security securityHelper) {
        this.namespace = namespace;
        this.namespaceConfig = namespaceConfig;
        this.securityHelper = securityHelper;
    }
    
    
    // ----------------------------------------------------- Instance Variables
    
    
    /**
     * Namespace.
     */
    private Namespace namespace;
    
    
    /**
     * Namespace configuration.
     */
    private NamespaceConfig namespaceConfig;
    
    
    /**
     * Security helper.
     */
    private Security securityHelper;
    
    
    // ----------------------------------------------------------- Lock Methods
    
    
    /**
     * Put a lock on a subject.
     * 
     * @param crendentialsToken Credentials token
     * @param token Object containing all the lock information
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectNotFoundException One of the objects referenced 
     * in the lock token were not found
     * @exception ObjectIsAlreadyLockedException Object is already locked 
     * with an incompatible lock token
     * @exception AccessDeniedException Insufficient credentials to allow 
     * object locking
     */
    public void lock(SlideToken slideToken, NodeLock token)
        throws ServiceAccessException, ObjectIsAlreadyLockedException, 
        AccessDeniedException, ObjectNotFoundException {
        
        ObjectIsAlreadyLockedException nestedException = 
            new ObjectIsAlreadyLockedException(token.getObjectUri());
        boolean canLock = 
            !isLockedInternal(slideToken, token, true, nestedException);
        
        // Then we try to lock the subject.
        // If the User doesn't have enough priviledges to accomplish this 
        // action, we will get a SecurityException which will in turn be 
        // thrown by this function.
        if (canLock) {
            Uri objectUri = namespace.getUri(slideToken, token.getObjectUri());
            ObjectNode lockedObject = objectUri.getStore()
                .retrieveObject(objectUri);
            securityHelper
                .checkCredentials(slideToken, lockedObject, 
                                  namespaceConfig.getLockObjectAction());
            Uri lockedUri = namespace.getUri(slideToken, token.getObjectUri());
            lockedUri.getStore().putLock(lockedUri, token);
        } else {
            throw nestedException;
        }
	
    }
    
    
    /**
     * Removes a lock.
     * 
     * @param crendentialsToken Credentials token
     * @param token Object containing all the lock information
     * @exception ServiceAccessException Low level service access exception
     * @exception LockTokenNotFoundException Cannot find the Lock in the 
     * Lock Store service
     */
    public void unlock(SlideToken slideToken, NodeLock token)
        throws ServiceAccessException, LockTokenNotFoundException {
        try {
            ObjectNode principal = securityHelper.getPrincipal(slideToken);
            
            // the current principal is the owner of the lock
            boolean condition1 = 
                token.getSubjectUri().startsWith(principal.getUri());
            // Are lock tokens enforced ?
            boolean condition2 = slideToken.isEnforceLockTokens();
            // Was a proper lock token given out ?
            boolean condition3 = checkLockToken(slideToken, token);
            
            if ((!condition1 && !condition2) || (!condition3 && condition2)) {
                return;
            }
        } catch (ObjectNotFoundException e) {
            return;
        }
        Uri lockedUri = namespace.getUri(slideToken, token.getObjectUri());
        lockedUri.getStore().removeLock(lockedUri, token);
    }
    
    
    /**
     * Removes a set of linked locks.
     * 
     * @param crendentialsToken Credentials token
     * @exception ServiceAccessException Low level service access exception
     * @exception LockTokenNotFoundException Cannot find the Lock in the 
     * Lock Store service
     * @exception ObjectNotFoundException One of the objects referenced 
     * in the lock token were not found
     */
    public void unlock(SlideToken slideToken, String objectUri, 
                       String lockId)
        throws ServiceAccessException, LockTokenNotFoundException, 
        ObjectNotFoundException {
        
        Enumeration locksList = enumerateLocks(slideToken, objectUri);
        while (locksList.hasMoreElements()) {
            NodeLock currentLock = (NodeLock) locksList.nextElement();
            if (currentLock.getLockId().equals(lockId)) {
                if (slideToken.isEnforceLockTokens())
                    slideToken.addLockToken(lockId);
                unlock(slideToken, currentLock);
            }
        }
        
    }
    
    
    /**
     * Renew a lock.
     * 
     * @param crendentialsToken Credentials token
     * @param token Object containing the lock info, as well as the new 
     * desired expiration date
     * @exception ServiceAccessException Low level service access exception
     * @exception LockTokenNotFoundException Cannot find the Lock in 
     * the Lock Store service
     */
    public void renew(SlideToken slideToken, NodeLock token, 
                      Date newExpirationDate)
        throws ServiceAccessException, LockTokenNotFoundException {
        token.setExpirationDate(newExpirationDate);
        Uri lockedUri = namespace.getUri(slideToken, token.getObjectUri());
        lockedUri.getStore().renewLock(lockedUri, token);
    }
    
    
    /**
     * Renew a set of linked locks.
     * 
     * @param crendentialsToken Credentials token
     * @exception ServiceAccessException Low level service access exception
     * @exception LockTokenNotFoundException Cannot find the Lock in the 
     * Lock Store service
     * @exception ObjectNotFoundException One of the objects referenced 
     * in the lock token were not found
     */
    public void renew(SlideToken slideToken, String objectUri, 
                      String lockId, Date newExpirationDate)
        throws ServiceAccessException, LockTokenNotFoundException, 
        ObjectNotFoundException {
        
        Enumeration locksList = enumerateLocks(slideToken, objectUri);
        while (locksList.hasMoreElements()) {
            NodeLock currentLock = (NodeLock) locksList.nextElement();
            if (currentLock.getLockId().equals(lockId)) {
                renew(slideToken, currentLock, newExpirationDate);
            }
        }
        
    }
    
    
    /**
     * Kills locks.
     * 
     * @param crendentialsToken Credentials token
     * @param subject Subject on which locks are to be removed
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectNotFoundException One of the objects referenced in 
     * the lock token were not found
     * @exception LockTokenNotFoundException Cannot find the Lock in the 
     * Lock Store service
     * @exception AccessDeniedException Insufficient credentials to allow 
     * forced removal of locks
     */
    public void kill(SlideToken slideToken, SubjectNode subject)
        throws ServiceAccessException, AccessDeniedException, 
        LockTokenNotFoundException, ObjectNotFoundException {
        
        // We retrieve the enumeration of locks which have been put on the 
        // subject.
        Uri subjectUri = namespace.getUri(slideToken, subject.getUri());
        Enumeration locks = subjectUri.getStore()
            .enumerateLocks(subjectUri);
        // Then, try to kill each individual lock.
        while (locks.hasMoreElements()) {
            securityHelper
                .checkCredentials(slideToken, subject, 
                                  namespaceConfig.getKillLockAction());
            subjectUri.getStore()
                .killLock(subjectUri, (NodeLock) locks.nextElement());
        }
        
    }
    
    
    /**
     * Enumerate locks on a subject.
     * 
     * @param slideToken Credentials token
     * @param objectUri Object on which we want to enumerate locks
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectNotFoundException One of the objects referenced 
     * in the lock token were not found
     */
    public Enumeration enumerateLocks(SlideToken slideToken, 
                                      String objectUri)
        throws ServiceAccessException, ObjectNotFoundException, 
        LockTokenNotFoundException {
        // We retrieve the LockStore service from the namespace.
        Vector locksVector = new Vector();
        
        Uri subjectUri = namespace.getUri(slideToken, objectUri);
        Enumeration scopes = subjectUri.getScopes();
        
        while (scopes.hasMoreElements()) {
            String currentScope = (String) scopes.nextElement();
            Uri currentScopeUri = namespace.getUri(slideToken, currentScope);
            Enumeration currentLocks = 
                currentScopeUri.getStore()
                .enumerateLocks(currentScopeUri);
            while (currentLocks.hasMoreElements()) {
                NodeLock currentLockToken = 
                    (NodeLock) currentLocks.nextElement();
                if (currentLockToken.hasExpired()) {
                    currentScopeUri.getStore()
                        .removeLock(currentScopeUri, currentLockToken);
                } else {
                    locksVector.addElement(currentLockToken);
                }
            }
        }
        
        return locksVector.elements();
    }
    
    
    /**
     * Utility function for lock checking. Mirrors Security.checkCredentials.
     * 
     * @param token Credetials token
     * @param subject Object on which the action is performed
     * @param action Action to test
     * @exception ObjectLockedException Can't perform specified action 
     * on object
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectNotFoundException One of the objects referenced 
     * in the lock token were not found
     */
    public void checkLock(SlideToken token, 
                          ObjectNode object, ActionNode action)
        throws ServiceAccessException, ObjectNotFoundException, 
        ObjectLockedException {
        
        if (Configuration.useIntegratedLocking()) {
            
            Uri objectUri = namespace.getUri(token, object.getUri());
            ObjectNode realObject = objectUri.getStore()
                .retrieveObject(objectUri);
            Uri subjectUri = null;
            subjectUri = namespace.getUri
                (token, namespaceConfig.getUsersPath() + "/"
                 + token.getCredentialsToken().getPublicCredentials());
            SubjectNode subject = (SubjectNode) subjectUri
                .getStore().retrieveObject(subjectUri);
            checkLock(token, realObject, subject, action);
            
        }
    }
    
    
    /**
     * Check locking for a specific action and credential.
     * 
     * @param slideToken Credetials token
     * @param subject Subject to test
     * @param user User to test
     * @param action Action to test
     * @exception ObjectLockedException Can't perform specified action 
     * on object
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectNotFoundException One of the objects referenced 
     * in the lock token were not found
     */
    public void checkLock(SlideToken slideToken, 
                          ObjectNode subject, SubjectNode user, 
                          ActionNode action)
        throws ServiceAccessException, ObjectNotFoundException, 
        ObjectLockedException {
        
        if (Configuration.useIntegratedLocking()) {
            if (isLocked(slideToken, subject, user, action, false)) {
                throw new ObjectLockedException(subject.getUri());
            }
        }
        
    }
    
    
    /**
     * Tests if an element is locked.
     * 
     * @param crendentialsToken Credentials token
     * @param subject Subject to test
     * @param user User to test
     * @param action Locked for action
     * @return boolean True if the subject is locked
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectNotFoundException One of the objects referenced 
     * in the lock token were not found
     */
    public boolean isLocked(SlideToken slideToken, 
                            ObjectNode subject, SubjectNode user, 
                            ActionNode action, boolean tryToLock)
        throws ServiceAccessException, ObjectNotFoundException {
        NodeLock token = new NodeLock(subject, user, action, 
                                      new Date(), false);
        return isLocked(slideToken, token, tryToLock);
    }
    
    
    /**
     * Tests if an element is locked.
     * 
     * @param crendentialsToken Credentials token
     * @param subject Subject to test
     * @param user User to test
     * @param action Locked for action
     * @param inheritance Set to true if we want to check if any children 
     * is locked with an incompatible lock
     * @param tryToLock True if the check is intended to check whether 
     * or not we can put a new lock
     * @return boolean True if the subject is locked
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectNotFoundException One of the objects referenced 
     * in the lock token were not found
     */
    public boolean isLocked(SlideToken slideToken, 
                            ObjectNode subject, SubjectNode user, 
                            ActionNode action, boolean inheritance, 
                            boolean tryToLock)
        throws ServiceAccessException, ObjectNotFoundException {
        NodeLock token = new NodeLock(subject, user, action, new Date(), 
                                      inheritance);
        return isLocked(slideToken, token, tryToLock);
    }
    
    
    /**
     * Tests if an element is locked.
     * 
     * @param token Lock token to test
     * @return boolean True if locked
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectNotFoundException One of the objects referenced 
     * in the lock token were not found
     */
    public boolean isLocked(SlideToken slideToken, NodeLock token,
                            boolean tryToLock)
        throws ServiceAccessException, ObjectNotFoundException {
        
        return isLockedInternal(slideToken, token, tryToLock, null);
        
    }
    
    
    // -------------------------------------------------------- Private Methods
    
    
    /**
     * Tests if an element is locked.
     * 
     * @param token Lock token to test
     * @return boolean True if locked
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectNotFoundException One of the objects referenced 
     * in the lock token were not found
     */
    private boolean isLockedInternal
        (SlideToken slideToken, NodeLock token,
         boolean tryToLock, ObjectIsAlreadyLockedException nestedException)
        throws ServiceAccessException, ObjectNotFoundException {

        Uri objectUri = namespace.getUri(slideToken, token.getObjectUri());
        ObjectNode initialObject = objectUri.getStore()
            .retrieveObject(objectUri);
        Enumeration scopes = objectUri.getScopes();
        
        // At the end of the test, this boolean's value is true if we can
        // actually put the lock on the desired subject.
        boolean isLocked = false;
        
        // We parse all of the scopes which encompass the subject we want
        // to lock.
        // First, we parse all the parents of the subject.
        while (!isLocked && scopes.hasMoreElements()) {
            String currentScope = (String) scopes.nextElement();
            Uri currentScopeUri = namespace.getUri(slideToken, currentScope);
            Enumeration locks = currentScopeUri.getStore()
                .enumerateLocks(currentScopeUri);
            
            while (locks.hasMoreElements()) {
                NodeLock currentLockToken = (NodeLock) locks.nextElement();
                if (!isCompatible(slideToken, token, currentLockToken, 
                                  tryToLock)) {
                    isLocked = true;
                    if (nestedException != null) {
                        nestedException.addException
                            (new ObjectLockedException
                             (currentScopeUri.toString()));
                    }
                }
            }
        }
        
        // Then, if the desired scope is inheritable, we parse the 
        // locked subject's children to see if any of them has been 
        // locked with an incompatible lock.
        if (token.isInheritable()) {
            Stack childrenStack = new Stack();
            childrenStack.push(initialObject);
            while (!isLocked && !childrenStack.empty()) {
                ObjectNode currentObject = (ObjectNode) childrenStack.pop();
                Uri currentObjectUri = 
                    namespace.getUri(slideToken, currentObject.getUri());
                // We test the compatibility of the child
                Enumeration locks = currentObjectUri.getStore()
                    .enumerateLocks(currentObjectUri);
                
                while (locks.hasMoreElements()) {
                    NodeLock currentLockToken = (NodeLock) locks.nextElement();
                    if (!isCompatible(slideToken, token, 
                                      currentLockToken, tryToLock)) {
                        isLocked = true;
                        if (nestedException != null) {
                            nestedException.addException
                                (new ObjectLockedException
                                 (currentObjectUri.toString()));
                        }
                    }
                }
                
                // We get the children and add them to the Stack.
                
                Vector childrenVector = new Vector();
                Enumeration childrenUri = currentObject.enumerateChildren();
                while (childrenUri.hasMoreElements()) {
                    String childUri = (String) childrenUri.nextElement();
                    Uri tempUri = namespace.getUri(slideToken, childUri);
                    ObjectNode child = tempUri.getStore()
                        .retrieveObject(tempUri);
                    childrenVector.addElement(child);
                }
                
                Enumeration children = childrenVector.elements();
                while (children.hasMoreElements()) {
                    ObjectNode tempObject = 
                        (ObjectNode) children.nextElement();
                    childrenStack.push(tempObject);
                }
            }
        }
        
        return isLocked;
        
    }
    
    
    /**
     * Return true if a lock token for this lock has been given in the 
     * credentials token.
     * 
     * @param slideToken Current credentials token
     * @param token Lock token
     */
    private boolean checkLockToken(SlideToken slideToken, 
                                   NodeLock token) {
        
        if (!slideToken.isEnforceLockTokens())
            return true;
        
        return (slideToken.checkLockToken(token.getLockId()));
        
    }
    
    
    /**
     * Tells whether or not two locks are compatible.
     * 
     * @param token1 First token
     * @param token2 Second token : object's lock
     * @param tryToLock True if we want to check for a lock creation
     * @return boolean True if the locks are compatible
     */
    private boolean isCompatible(SlideToken slideToken,
                                 NodeLock token1, NodeLock token2, 
                                 boolean tryToLock) {
        /*
          System.out.println("**** Check lock ****");
          System.out.println("Lock 1 : " + token1.getSubjectUri() + " action " 
          + token1.getTypeUri());
          System.out.println("Lock 2 : " + token2.getSubjectUri() + " on "
          + token2.getObjectUri() + " action " 
          + token2.getTypeUri());
        */
        
        boolean compatible = true;
        
        // We first check whether or not the lock is still valid
        if (token2.hasExpired()) {
            // Since the lock has expired, it is removed
            try {
                Uri token2Uri = namespace.getUri(slideToken, 
                                                 token2.getObjectUri());
                token2Uri.getStore().removeLock(token2Uri, token2);
            } catch (SlideException e) {
                e.printStackTrace();
            }
        } else {
            // The lockTypes are not equal
            boolean condition1 = (tryToLock) 
                && (token2.getTypeUri().equals(token1.getTypeUri()));
            // The user tested is not parent of the token's user
            boolean condition2 = 
                (token1.getSubjectUri().startsWith(token2.getSubjectUri()));
            // The two lock types are the same
            boolean condition3 = token2.getTypeUri()
                .startsWith(token1.getTypeUri());
            // The lockType is exclusive
            boolean condition4 = token1.isExclusive();
            // Was a proper lock token given out ?
            boolean condition5 = checkLockToken(slideToken, token2);
            // Are lock tokens enforced ?
            boolean condition6 = slideToken.isEnforceLockTokens();
            if ((condition1 && condition4)
                || (condition3 && !condition2 && !condition6) 
                || (condition3 && !condition5 && condition6)) {
                compatible = false;
            }
            /*
              System.out.println("C1:" + condition1);
              System.out.println("C2:" + condition2);
              System.out.println("C3:" + condition3);
              System.out.println("C4:" + condition4);
              System.out.println("C5:" + condition5);
              System.out.println("C6:" + condition6);
              System.out.println("Compatible:" + compatible);
            */
        }
        return compatible;
    }
    
}
