/*
 * $Header: /home/cvs/jakarta-slide/src/share/org/apache/slide/structure/Structure.java,v 1.7 2001/05/29 02:42:43 remm Exp $
 * $Revision: 1.7 $
 * $Date: 2001/05/29 02:42:43 $
 *
 * ====================================================================
 *
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:  
 *       "This product includes software developed by the 
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Tomcat", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written 
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * [Additional notices, if required by prior licensing conditions]
 *
 */ 

package org.apache.slide.structure;

import java.io.*;
import java.util.Enumeration;
import java.util.Stack;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Date;
import org.apache.slide.common.*;
import org.apache.slide.lock.*;
import org.apache.slide.security.*;
import org.apache.slide.authenticate.CredentialsToken;

/**
 * Structure helper.
 * 
 * @author <a href="mailto:remm@apache.org">Remy Maucherat</a>
 * @version $Revision: 1.7 $
 */
public interface Structure {
    
    
    // ------------------------------------------------------ Interface Methods
    
    
    /**
     * Get object's children.
     * 
     * @param token Credentials token
     * @param object Slide object
     * @return Enumeration Enumeration of children objects
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectNotFoundException One of the children specified 
     * by the object were not found
     * @exception LinkedObjectNotFoundException Cannot happen
     */
    Enumeration getChildren(SlideToken token, ObjectNode object) 
        throws ServiceAccessException, ObjectNotFoundException, 
        LinkedObjectNotFoundException;
    
    
    /**
     * Get object's parent.
     * 
     * @param token Credentials token
     * @param object Slide object
     * @return ObjectNode Object's parent; null if the object specified
     * if the root node of the namespace
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectNotFoundException The parent object specified 
     * by the object was not found
     * @exception LinkedObjectNotFoundException Cannot happen : a link 
     * cannot have children
     * @exception AccessDeniedException Credentials token does not 
     * have permission to perform the action
     */
    ObjectNode getParent(SlideToken token, ObjectNode object) 
        throws ServiceAccessException, ObjectNotFoundException, 
        LinkedObjectNotFoundException, AccessDeniedException;
    
    
    /**
     * Get an object, following any link.
     * 
     * @param token Credentials token
     * @param strUri Object uri
     * @return ObjectNode Object specified by the Uri
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectNotFoundException The given Uri doesn't have 
     * any matching ObjectNode
     * @exception LinkedObjectNotFoundException A link object encountered 
     * during Uri resolution
     * doesn't have a valid target
     * @exception AccessDeniedException Credentials token does not have 
     * permission to perform the action
     */
    ObjectNode retrieve(SlideToken token, String strUri) 
        throws ServiceAccessException, ObjectNotFoundException, 
        LinkedObjectNotFoundException, AccessDeniedException;
    
    
    /**
     * Get an object, following any link.
     * 
     * @param token Credentials token
     * @param strUri Object uri
     * @param translateLastUriElement If set to true and if the uri is 
     * associated with a link, the function will return the linked object. 
     * Otherwise, the link object is returned
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectNotFoundException The given Uri doesn't have any 
     * matching ObjectNode
     * @exception LinkedObjectNotFoundException A link object encountered 
     * during Uri resolution doesn't have a valid target
     * @exception AccessDeniedException Credentials token does not have 
     * permission to perform the action
     */
    ObjectNode retrieve(SlideToken token, String strUri, 
                        boolean translateLastUriElement) 
        throws ServiceAccessException, ObjectNotFoundException, 
        LinkedObjectNotFoundException, AccessDeniedException;
    
    
    /**
     * Create an object.
     * 
     * @param token Credentials token
     * @param object Object we want to create
     * @param uri Location in the namespace where we want to create the object
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectAlreadyExistException An object already exist 
     * at the specified Uri
     * @exception ObjectNotFoundException Update of the parent object 
     * failed because the parent object does no longer exist. Should not 
     * happen, and indicate a critical error
     * @exception LinkedObjectNotFoundException Retrieval of a link target 
     * failed during Uri resolution
     * @exception AccessDeniedException Credentials token does not have 
     * permission to perform the action
     */
    void create(SlideToken token, ObjectNode object, String strUri)
        throws ServiceAccessException, ObjectAlreadyExistsException, 
        ObjectNotFoundException, LinkedObjectNotFoundException, 
        AccessDeniedException;
    
    
    /**
     * Create a link.
     * 
     * @param token Credentials token
     * @param link Link object we want to create
     * @param linkUri Location in the namespace where we want to create 
     * the object
     * @param linkedObject Link target
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectAlreadyExistException An object already exist 
     * at the specified Uri
     * @exception ObjectNotFoundException Update of the parent object 
     * failed because the parent object does no longer exist. Should not 
     * happen, and indicate a critical error
     * @exception LinkedObjectNotFoundException Retrieval of a link target 
     * failed during Uri resolution
     * @exception AccessDeniedException Credentials token does not have 
     * permission to perform the action
     */
    void createLink(SlideToken token, LinkNode link, String linkUri, 
                    ObjectNode linkedObject)
        throws ServiceAccessException, ObjectAlreadyExistsException, 
        ObjectNotFoundException, LinkedObjectNotFoundException, 
        AccessDeniedException;
    
    
    /**
     * Store (ie, update) an object.
     * 
     * @param token Credentials token
     * @param object Object to update
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectNotFoundException The update failed because one 
     * object was not found during Uri resolution
     * @exception LinkedObjectNotFoundException Retrieval of a link target 
     * failed during Uri resolution
     * @exception AccessDeniedException Credentials token does not have 
     * permission to perform the action
     */
    void store(SlideToken token, ObjectNode object)
        throws ServiceAccessException, ObjectNotFoundException, 
        AccessDeniedException, LinkedObjectNotFoundException;
    
    
    /**
     * Remove an object.
     * 
     * @param token Credentials token
     * @param object Slide object to remove
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectNotFoundException The update failed because one 
     * object was not found during Uri resolution
     * @exception ObjectHasChildrenException Removal failed because object 
     * has children
     * @exception LinkedObjectNotFoundException Retrieval of a link target 
     * failed during Uri resolution
     * @exception AccessDeniedException Credentials token does not have 
     * permission to perform the action
     */
    void remove(SlideToken token, ObjectNode object)
        throws ServiceAccessException, ObjectNotFoundException, 
        ObjectHasChildrenException, AccessDeniedException, 
        LinkedObjectNotFoundException, ObjectLockedException;
    
}
