/*
 * $Header: /home/cvs/jakarta-slide/src/share/org/apache/slide/structure/StructureImpl.java,v 1.21 2001/07/11 17:18:10 remm Exp $
 * $Revision: 1.21 $
 * $Date: 2001/07/11 17:18:10 $
 *
 * ====================================================================
 *
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Tomcat", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * [Additional notices, if required by prior licensing conditions]
 *
 */

package org.apache.slide.structure;

import java.io.*;
import java.util.Enumeration;
import java.util.Stack;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Date;
import org.apache.slide.common.*;
import org.apache.slide.lock.*;
import org.apache.slide.security.*;
import org.apache.slide.authenticate.CredentialsToken;

/**
 * Data helper class.
 *
 * @author <a href="mailto:remm@apache.org">Remy Maucherat</a>
 * @version $Revision: 1.21 $
 */
public final class StructureImpl implements Structure {
    
    
    // ----------------------------------------------------------- Constructors
    
    
    /**
     * Constructor.
     *
     * @param namespace Namespace associated with the helper object
     * @param namespaceConfig Configuration of the namespace
     * @param securityHelper Associated security helper
     */
    public StructureImpl(Namespace namespace, NamespaceConfig namespaceConfig,
                         Security securityHelper, Lock lockHelper) {
        this.namespace = namespace;
        this.namespaceConfig = namespaceConfig;
        this.securityHelper = securityHelper;
        this.lockHelper = lockHelper;
    }
    
    
    // ----------------------------------------------------- Instance Variables
    
    
    /**
     * Namespace.
     */
    private Namespace namespace;
    
    
    /**
     * Namespace configuration.
     */
    private NamespaceConfig namespaceConfig;
    
    
    /**
     * Security helper.
     */
    private Security securityHelper;
    
    
    /**
     * Lock helper.
     */
    private Lock lockHelper;
    
    
    // ------------------------------------------------------ Structure Methods
    
    
    /**
     * Get object's children.
     *
     * @param token Credentials token
     * @param object Slide object
     * @return Enumeration Enumeration of children objects
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectNotFoundException One of the children specified by
     * the object were not found
     * @exception LinkedObjectNotFoundException Cannot happen
     */
    public Enumeration getChildren(SlideToken token, ObjectNode object)
        throws ServiceAccessException, ObjectNotFoundException,
        LinkedObjectNotFoundException {
        Enumeration childrenUri = object.enumerateChildren();
        Vector result = new Vector();
        while (childrenUri.hasMoreElements()) {
            String childUri = (String) childrenUri.nextElement();
            try {
                ObjectNode child = retrieve(token, childUri, false);
                result.addElement(child);
            } catch (AccessDeniedException e) {
            }
        }
        return result.elements();
    }
    
    
    /**
     * Get object's parent.
     *
     * @param token Credentials token
     * @param object Slide object
     * @return ObjectNode Object's parent; null if the object specified
     * if the root node of the namespace
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectNotFoundException The parent object specified by
     * the object was not found
     * @exception LinkedObjectNotFoundException Cannot happen : a link cannot
     * have children
     * @exception AccessDeniedException Credentials token does not have
     * permission to perform the action
     */
    public ObjectNode getParent(SlideToken token, ObjectNode object)
        throws ServiceAccessException, ObjectNotFoundException,
        LinkedObjectNotFoundException, AccessDeniedException {
        String objectUriStr = object.getUri();
        Uri parentUri = namespace.getUri(token, objectUriStr).getParentUri();
        if (parentUri == null) {
            return null;
        }
        String parentUriStr = parentUri.toString();
        ObjectNode parent = retrieve(token, parentUriStr);
        return parent;
    }
    
    
    /**
     * Get an object, following any link.
     *
     * @param token Credentials token
     * @param strUri Object uri
     * @return ObjectNode Object specified by the Uri
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectNotFoundException The given Uri doesn't have
     * any matching ObjectNode
     * @exception LinkedObjectNotFoundException A link object encountered
     * during Uri resolution doesn't have a valid target
     * @exception AccessDeniedException Credentials token does not have
     * permission to perform the action
     */
    public ObjectNode retrieve(SlideToken token, String strUri)
        throws ServiceAccessException, ObjectNotFoundException,
        LinkedObjectNotFoundException, AccessDeniedException {
        return retrieve(token, strUri, true);
    }
    
    
    /**
     * Get an object, following any link.
     *
     * @param token Credentials token
     * @param strUri Object uri
     * @param translateLastUriElement If set to true and if the uri
     * is associated with a link, the function will return the linked object.
     * Otherwise, the link object is returned
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectNotFoundException The given Uri doesn't have any
     * matching ObjectNode
     * @exception LinkedObjectNotFoundException A link object encountered
     * during Uri resolution doesn't have a valid target
     * @exception AccessDeniedException Credentials token does not have
     * permission to perform the action
     */
    public ObjectNode retrieve(SlideToken token, String strUri,
                               boolean translateLastUriElement)
        throws ServiceAccessException, ObjectNotFoundException,
        LinkedObjectNotFoundException, AccessDeniedException {
        
        Uri uri = namespace.getUri(token, strUri);
        
        ObjectNode result = null;
        
        // First of all, we try to load the object directly from the given Uri.
        try {
            result = uri.getStore().retrieveObject(uri);
            securityHelper.checkCredentials
                (token, result, namespaceConfig.getReadObjectAction());
            if ((translateLastUriElement) && (result instanceof LinkNode)) {
                LinkNode link = (LinkNode) result;
                Uri linkedUri = namespace.getUri(token, link.getLinkedUri());
                result = linkedUri.getStore()
                    .retrieveObject(linkedUri);
                securityHelper.checkCredentials
                    (token, result, namespaceConfig.getReadObjectAction());
            }
        } catch (ObjectNotFoundException e) {
        }
        
        // If the attempt to load the uri failed, it means there is at least
        // one link in the uri (or that the uri doe'sn't have any associated
        // object).
        if (result == null) {
            
            String resolvedUri = uri.toString();
            
            // 1 - Tokemization of the Uri
            UriTokenizer uriTokenizer = new UriTokenizer(uri.getNamespace(),
                                                         resolvedUri);
            
            // 2 - For each element of the Uri
            Uri courUri = null;
            ObjectNode courObject = null;
            while (uriTokenizer.hasMoreElements()) {
                
                // 3 - Load object's class from the uri. If the object
                // does not exist, a DataException is thrown.
                courUri = uriTokenizer.nextUri();
                courObject = courUri.getStore()
                    .retrieveObject(courUri);
                
                // We check to see if the credentials gives access to
                //the current object
                securityHelper.checkCredentials
                    (token, courObject, namespaceConfig.getReadObjectAction());
                
                // 4 - Test if object is a link, ie if it is an instance
                // of LinkNode or one of its subclasses
                if (((translateLastUriElement)
                     && (courObject instanceof LinkNode)) ||
                    ((!translateLastUriElement)
                     && (uriTokenizer.hasMoreElements())
                     && (courObject instanceof LinkNode))
                    ) {
                    
                    // 5 - If the object is a link, we get the uri of
                    // the linked object
                    // Note : courUri still IS the Uri of the link, and so,
                    // in a way courUri is the parent of linkedUri.
                    Uri linkedUri = namespace
                        .getUri(((LinkNode) courObject).getLinkedUri());
                    
                    // 6 - We replace the courUri scope in the original uri
                    String courStrUri = courUri.toString();
                    resolvedUri = linkedUri.toString()
                        + resolvedUri.substring(courStrUri.length());
                    
                    // 7 - We tokenize again the uri
                    uriTokenizer = new UriTokenizer(uri.getNamespace(),
                                                    resolvedUri);
                    
                    // 8 - We parse it till we get back to the point
                    // where we stopped
                    boolean isUriFound = false;
                    while ((!isUriFound) && (uriTokenizer.hasMoreElements())) {
                        if (linkedUri.equals(uriTokenizer.nextUri())) {
                            isUriFound = true;
                        }
                    }
                    if (!isUriFound) {
                        throw new LinkedObjectNotFoundException(courUri,
                                                                resolvedUri);
                    }
                    
                }
                
                // 9 - We continue to go down in the Uri tree
            }
            
            // 10 - We return the last object which has been found
            
            result = courObject;
        }
        
        return result;
        
    }
    
    
    /**
     * Create an object.
     *
     * @param token Credentials token
     * @param object Object we want to create
     * @param uri Location in the namespace where we want to create the object
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectAlreadyExistException An object already exist at
     * the specified Uri
     * @exception ObjectNotFoundException Update of the parent object
     * failed because the parent object does no longer exist. Should not
     * happen, and indicate a critical error
     * @exception LinkedObjectNotFoundException Retrieval of a link target
     * failed during Uri resolution
     * @exception AccessDeniedException Credentials token does not have
     * permission to perform the action
     */
    public void create(SlideToken token, ObjectNode object,
                       String strUri)
        throws ServiceAccessException, ObjectAlreadyExistsException,
        ObjectNotFoundException, LinkedObjectNotFoundException,
        AccessDeniedException {
        
        // Checking roles
        Enumeration roles = securityHelper.getRoles(object);
        while (roles.hasMoreElements()) {
            if (!securityHelper.hasRole(token, (String)roles.nextElement())) {
                // Allow only the namespace admin to create roles 
                // he doesn't have
                Uri rootUri = namespace.getUri(token, "/");
                ObjectNode rootObject =
                    rootUri.getStore().retrieveObject(rootUri);
                securityHelper.checkCredentials
                    (token, rootObject,
                     namespaceConfig.getGrantPermissionAction());
                break;
            }
        }
        
        String resolvedUri = strUri;
        
        // 1 - Tokenization of the Uri
        UriTokenizer uriTokenizer = new UriTokenizer(namespace, resolvedUri);
        
        // 2 - For each element of the Uri
        Uri courUri = null;
        ObjectNode courObject = null;
        ObjectNode parentObject = null;
        
        boolean alreadyExists = false;
        
        while (uriTokenizer.hasMoreElements()) {
            
            parentObject = courObject;
            
            // 3 - Load object's class from the uri. If the object does
            // not exist, a DataException is thrown.
            courUri = uriTokenizer.nextUri();
            try {
                courObject = courUri.getStore()
                    .retrieveObject(courUri);
                securityHelper
                    .checkCredentials(token, courObject,
                                      namespaceConfig.getReadObjectAction());
                if (!uriTokenizer.hasMoreElements()) {
                    // The object already exists
                    alreadyExists = true;
                }
            } catch (ObjectNotFoundException e) {
                // Load failed, probably because object was not found
                // We try to create a new one.
                // We have to test if the uri is the last in the list,
                // we must create the requested element.
                // By default, we create a SubjectNode.
                ObjectNode newObject = null;
                if (uriTokenizer.hasMoreElements()) {
                    throw new ObjectNotFoundException(courUri);
                } else {
                    newObject = object;
                }
                if (parentObject != null) {
                    securityHelper
                        .checkCredentials(token, courObject, namespaceConfig
                                          .getCreateObjectAction());
                    
                    // The role who creates the new object should have
                    // read/write access permissions
                    // granted on the newly created object.
                    // In fact, everything but modify permssions.
                    // If a permission is given, it is given as a
                    // non heritable permission
                    // - Read properties
                    // - Modify revisions
                    // - Delete object
                    // - Put lock
                    //
                    // Quite possibly in the future, it will be possible
                    // to define the default behavior for newly created
                    // object in the configuration files.
                    // This would then be the default behavior.
                    NodePermission grantedPermission = null;
                    String userUri = namespaceConfig.getUsersPath() + "/"
                        + token.getCredentialsToken().getPublicCredentials();
                    Domain.debug("Checking basic permissions on new object");
                    /*
                      try {
                      securityHelper.checkCredentials
                      (token, courObject,
                      namespaceConfig.getReadPropertiesAction());
                      } catch (AccessDeniedException excep) {
                      grantedPermission = new NodePermission
                      (courUri.toString(), userUri,
                      namespaceConfig.getReadPropertiesAction().getUri(),
                      false);
                      securityHelper.grantPermission ...
                      //newObject.addPermission(grantedPermission);
                      }
                      try {
                      securityHelper.checkCredentials
                      (token, courObject,
                      namespaceConfig.getModifyRevisionsAction());
                      } catch (AccessDeniedException excep) {
                      grantedPermission = new NodePermission
                      (courUri.toString(), userUri,
                      namespaceConfig.getModifyRevisionsAction().getUri(),
                      false);
                      newObject.addPermission(grantedPermission);
                      }
                      try {
                      securityHelper.checkCredentials
                      (token, courObject,
                      namespaceConfig.getDeleteObjectAction());
                      } catch (AccessDeniedException excep) {
                      grantedPermission = new NodePermission
                      (courUri.toString(), userUri,
                      namespaceConfig.getDeleteObjectAction().getUri(), false);
                      newObject.addPermission(grantedPermission);
                      }
                      try {
                      securityHelper.checkCredentials
                      (token, courObject, namespaceConfig.getPutLockAction());
                      } catch (AccessDeniedException excep) {
                      grantedPermission = new NodePermission
                      (courUri.toString(), userUri,
                      namespaceConfig.getPutLockAction().getUri(), false);
                      newObject.addPermission(grantedPermission);
                      }
                    */
                    Domain.debug
                        ("Basic permissions granted for user "
                         + token.getCredentialsToken().getPublicCredentials());
                    
                    // Now creating the new object
                    newObject.setUri(courUri.toString());
                    courUri.getStore().createObject
                        (courUri, newObject);
                    
                    // Add the newly created object to its parent's
                    // children list
                    parentObject.addChild(newObject);
                    //namespace.getUri(token, parentObject.getUri())
                    //.getDataSource().storeObject(parentObject, false);
                    store(token, parentObject);
                    
                } else {
                    throw new ObjectNotFoundException(courUri);
                }
                courObject = newObject;
            }
            
            // 4 - Test if object is a link, ie if it is an instance of
            // LinkNode or one of its subclasses
            if ((uriTokenizer.hasMoreElements())
                && (courObject instanceof LinkNode)) {
                
                // 5 - If the object is a link, we get the uri of the
                // linked object
                // Note : courUri still IS the Uri of the link, and so,
                // in a way courUri is the parent of linkedUri.
                Uri linkedUri = namespace
                    .getUri(((LinkNode) courObject).getLinkedUri());
                
                // 6 - We replace the courUri scope in the original uri
                String courStrUri = courUri.toString();
                resolvedUri = linkedUri.toString()
                    + resolvedUri.substring(courStrUri.length());
                
                // 7 - We tokenize again the uri
                uriTokenizer = new UriTokenizer(namespace, resolvedUri);
                
                // 8 - We parse it till we get back to the point
                // where we stopped
                boolean isUriFound = false;
                while ((!isUriFound) && (uriTokenizer.hasMoreElements())) {
                    if (linkedUri.equals(uriTokenizer.nextUri())) {
                        isUriFound = true;
                    }
                }
                if (!isUriFound) {
                    throw new LinkedObjectNotFoundException
                        (courUri, resolvedUri);
                }
                
            }
            
            // 9 - We continue to go down in the Uri tree
        }
        
        if (alreadyExists) {
            // if the object already exists map it anyway into
            // the node hierarchy, to prevent loose of nodes
            // during start up
            parentObject.addChild(courObject);
            store(token, parentObject);
            throw new ObjectAlreadyExistsException(strUri);
        }
        
    }
    
    
    /**
     * Create a link.
     *
     * @param token Credentials token
     * @param link Link object we want to create
     * @param linkUri Location in the namespace where we want to create
     * the object
     * @param linkedObject Link target
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectAlreadyExistException An object already exist at
     * the specified Uri
     * @exception ObjectNotFoundException Update of the parent object
     * failed because the parent object does no longer exist. Should not
     * happen, and indicate a critical error
     * @exception LinkedObjectNotFoundException Retrieval of a link target
     * failed during Uri resolution
     * @exception AccessDeniedException Credentials token does not have
     * permission to perform the action
     */
    public void createLink(SlideToken token, LinkNode link,
                           String linkUri, ObjectNode linkedObject)
        throws ServiceAccessException, ObjectAlreadyExistsException,
        ObjectNotFoundException, LinkedObjectNotFoundException,
        AccessDeniedException {
        link.setLinkedUri(linkedObject.getUri());
        create(token, link, linkUri);
    }
    
    
    /**
     * Store (ie, update) an object.
     *
     * @param token Credentials token
     * @param object Object to update
     * @param updatePermissions True to update permissions
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectNotFoundException The update failed because one
     * object was not found during Uri resolution
     * @exception LinkedObjectNotFoundException Retrieval of a link target
     * failed during Uri resolution
     * @exception AccessDeniedException Credentials token does not have
     * permission to perform the action
     */
    public void store(SlideToken token, ObjectNode object)
        throws ServiceAccessException, ObjectNotFoundException,
        AccessDeniedException, LinkedObjectNotFoundException {
        
        // Checking roles
        Enumeration roles = securityHelper.getRoles(object);
        while (roles.hasMoreElements()) {
            if (!securityHelper.hasRole(token, (String)roles.nextElement())) {
                // Allow only the namespace admin to create roles 
                // he doesn't have
                Uri rootUri = namespace.getUri(token, "/");
                ObjectNode rootObject =
                    rootUri.getStore().retrieveObject(rootUri);
                securityHelper.checkCredentials
                    (token, rootObject,
                     namespaceConfig.getGrantPermissionAction());
                break;
            }
        }
        
        ObjectNode realObject = retrieve(token, object.getUri(), false);
        securityHelper
            .checkCredentials(token, realObject,
                              namespaceConfig.getCreateObjectAction());
        Uri realObjectUri = namespace.getUri(token, realObject.getUri());
        realObjectUri.getStore().storeObject(realObjectUri, object);
        
    }
    
    
    /**
     * Remove an object.
     *
     * @param token Credentials token
     * @param object Slide object to remove
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectNotFoundException The update failed because
     * one object was not found during Uri resolution
     * @exception ObjectHasChildrenException Removal failed because object
     * has children
     * @exception LinkedObjectNotFoundException Retrieval of a link target
     * failed during Uri resolution
     * @exception AccessDeniedException Credentials token does not have
     * permission to perform the action
     */
    public void remove(SlideToken token, ObjectNode object)
        throws ServiceAccessException, ObjectNotFoundException,
        ObjectHasChildrenException, AccessDeniedException,
        LinkedObjectNotFoundException, ObjectLockedException {
        
        ObjectNode realObject = retrieve(token, object.getUri(), false);
        Enumeration enum = realObject.enumerateChildren();
        Uri uri = namespace.getUri(token, realObject.getUri());
        
        if (enum.hasMoreElements()) {
            throw new ObjectHasChildrenException(uri);
        }
        if (!object.getUri().equals("/")) {
            Uri curUri = namespace.getUri(token, realObject.getUri());
            Uri parentUri = curUri.getParentUri();
            ObjectNode obj = parentUri.getStore().retrieveObject(parentUri);
            obj.removeChild(realObject.getUri());
            securityHelper.checkCredentials
                (token, realObject, namespaceConfig.getRemoveObjectAction());
            lockHelper.checkLock
                (token, realObject, namespaceConfig.getRemoveObjectAction());
            store(token, obj);
            uri.getStore().removeObject(uri, realObject);
        }
    }
    
}
